/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef __plm_itk_h_
#define __plm_itk_h_

/* ITK 3.20 is missing this */
#if defined __cplusplus
#include <stddef.h>
#endif

#if (ITK_FOUND && !PLM_CUDA_COMPILE)
#include "itkConfigure.h"
#endif

#endif
