/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _bowtie_correction_h_
#define _bowtie_correction_h_

#include "plmreconstruct_config.h"

class Fdk_parms;
class Volume;

PLMRECONSTRUCT_C_API void bowtie_correction (Volume *vol, Fdk_parms *parms);


#endif
