/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _bragg_curve_h_
#define _bragg_curve_h_

#include "plmdose_config.h"

PLMDOSE_C_API
double
bragg_curve (
    double E_0,         /* in MeV */
    double sigma_E0,    /* in MeV */
    double z            /* in mm */
);

#endif
