/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _bspline_gm_h_
#define _bspline_gm_h_

#include "plmregister_config.h"

class Bspline_optimize;

PLMREGISTER_API void bspline_score_gm (Bspline_optimize *bod);

#endif
