/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _simplify_points_h_
#define _simplify_points_h_

#include "plmutil_config.h"

class Rt_study;

PLMUTIL_C_API void do_simplify (Rt_study *rtds, float percentage);

#endif
