/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _ss_img_stats_h_
#define _ss_img_stats_h_

#include "plmutil_config.h"
#include "itk_image_type.h"

PLMUTIL_C_API void ss_img_stats (UCharVecImageType::Pointer img);

#endif
