/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _gaussian_h_
#define _gaussian_h_

#include "plmbase_config.h"

PLMBASE_API float* create_ker (float coeff, int half_width);
PLMBASE_API void validate_filter_widths (int *fw_out, int *fw_in);
PLMBASE_API void kernel_stats (float* kerx, float* kery, float* kerz, int fw[]);

#endif
