/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _demon_opts_h_
#define _demon_opts_h_

#include "demons.h"

class Demons_options {
public:
    char* fixed_fn;
    char* moving_fn;
    char* output_vf_fn;
    char* output_img_fn;
    Demons_parms parms;
};

void parse_args (Demons_options* options, int argc, char* argv[]);

#endif
