/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _itk_image_header_compare_h_
#define _itk_image_header_compare_h_

#include "plmbase_config.h"

/* Return true if the headers are the same */
template<class T, class U> PLMBASE_API bool itk_image_header_compare (T image1, U image2);

#endif
