/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _groupwise_parms_h_
#define _groupwise_parms_h_

#include "plmregister_config.h"

class PLMREGISTER_API Groupwise_parms {
public:
    Groupwise_parms ();
    ~Groupwise_parms ();

public:
};

#endif
