/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _proj_image_filter_h_
#define _proj_image_filter_h_

#include "plmutil_config.h"

class Proj_image;

PLMUTIL_C_API void proj_image_filter (Proj_image *proj);

#endif
