/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _itk_union_h_
#define _itk_union_h_

#include "plmutil_config.h"
#include "itk_image_type.h"

PLMUTIL_API
UCharImageType::Pointer
itk_union (const UCharImageType::Pointer image_1,
    const UCharImageType::Pointer image_2);

#endif
