/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _pcmd_warp_h_
#define _pcmd_warp_h_

#include "plmcli_config.h"

class Warp_parms;

void
do_command_warp (int argc, char* argv[]);
void
warp_image_main (Warp_parms* parms);
void
warp_dij_main (Warp_parms* parms);
void
warp_pointset_main (Warp_parms* parms);

#endif
