/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _lua_tty_preview_h_
#define _lua_tty_preview_h_

#include "plmscript_config.h"

#if defined __cplusplus
extern "C" {
#endif

PLMSCRIPT_C_API int preview_portal (void* pli);

#if defined __cplusplus
}
#endif

#endif
