/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#include "plmbase_config.h"
#include <stdlib.h>
#include <stdio.h>

#include "plm_study.h"

class Plm_study_private {
public:

public:
    Plm_study_private () {
    }
    ~Plm_study_private () {
    }
};

Plm_study::Plm_study ()
{
}

Plm_study::~Plm_study ()
{
}

void
Plm_study::debug () const
{
}
