/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _xform_legacy_h_
#define _xform_legacy_h_

#include "plmbase_config.h"
#include <stdio.h>

class Xform;

PLMBASE_C_API void xform_legacy_load (Xform *xf, FILE* fp);

#endif
