/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _pcmd_autolabel_train_h_
#define _pcmd_autolabel_train_h_

#include "plmcli_config.h"

void do_command_autolabel_train (int argc, char *argv[]);

#endif
