/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _pcmd_script_h_
#define _pcmd_script_h_

#include "plmcli_config.h"

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

void do_command_script (int argc, char *argv[]);

#endif
