/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _pointset_warp_h_
#define _pointset_warp_h_

#include "plmbase_config.h"
#include "pointset.h"
#include "itk_image_type.h"

PLMBASE_API void
pointset_warp (
    Labeled_pointset *warped_pointset,
    Labeled_pointset *input_pointset,
    DeformationFieldType::Pointer vf);

#endif
