/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _bspline_optimize_lbfgsb_h_
#define _bspline_optimize_lbfgsb_h_

#include <stdio.h>

class Bspline_optimize;

PLMREGISTER_C_API void bspline_optimize_lbfgsb (Bspline_optimize *bod);

#endif
