/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _registration_resample_h_
#define _registration_resample_h_

#include "plmregister_config.h"
#include "stage_parms.h"
#include "volume.h"

Volume::Pointer
registration_resample_volume (
    const Volume::Pointer& vol,
    const Stage_parms* stage,
    const float resample_rate[3]
);

#endif
