/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _itk_align_center_h_
#define _itk_align_center_h_

#include "plmregister_config.h"
#include "xform.h"

class Registration_data;
class Stage_parms;

Xform::Pointer
do_itk_align_center (
    Registration_data* regd, const Xform::Pointer& xf_in, Stage_parms* stage);

Xform::Pointer
do_itk_align_center_of_gravity (
    Registration_data* regd, const Xform::Pointer& xf_in, Stage_parms* stage);

#endif
