#include "DlgGammaView.h"
#include "gamma_gui.h"

#define FIXME_BACKGROUND_MAX (-1200)

using namespace std;

DlgGammaView::DlgGammaView(): QDialog ()
{
    /* Sets up the GUI */
    ui.setupUi (this);
}

DlgGammaView::DlgGammaView(QWidget *parent): QDialog (parent)
{
    ui.setupUi (this);
    m_pParent = (gamma_gui*)(parent);  

}

DlgGammaView::~DlgGammaView()
{    
 
}