/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#include "plm_config.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

int
main (int argc, char *argv[])
{
    printf ("sizeof(int) = %d\n", sizeof(int));
    printf ("sizeof(long) = %d\n", sizeof(long));
    printf ("sizeof(void*) = %d\n", sizeof(void*));
    printf ("sizeof(size_t) = %d\n", sizeof(size_t));
    return 0;
}

