/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _fdk_opts_h_
#define _fdk_opts_h_

#include "plm_config.h"

class Fdk_parms;

#define OPTION_RESOLUTION_STRING "resolution"
#define OPTION_RESOLUTION 'r'
#define OPTION_INPUT_DIR_STRING "input-dir"
#define OPTION_INPUT_DIR 'I'
#define OPTION_OUTPUT_FILE_STRING "output-file"
#define OPTION_OUTPUT_FILE 'O'
#define OPTION_IMAGE_RANGE_STRING "image-range"
#define OPTION_IMAGE_RANGE 'a'
#define OPTION_SCALE_STRING "scale"
#define OPTION_SCALE 's'
#define OPTION_VOL_SIZE_STRING "volume-size"
#define OPTION_VOL_SIZE 'z'


void fdk_parse_args (Fdk_parms* parms, int argc, char* argv[]);


#endif
