/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _bspline_xform_legacy_h_
#define _bspline_xform_legacy_h_

#include "plmbase_config.h"
#include "bspline_xform.h"

PLMBASE_API Bspline_xform* 
bspline_xform_legacy_load (const char* filename);

#endif
