/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _autolabel_task_h_
#define _autolabel_task_h_

#include "plmsegment_config.h"
#include "dlib_trainer.h"

class Autolabel_task {
public:
    Autolabel_task () {
    }
public:
    enum Task_type {
        AUTOLABEL_TASK_X_POS,
        AUTOLABEL_TASK_Y_POS,
        AUTOLABEL_TASK_Z_POS
    };
public:
    Task_type task_type;
};

#endif
