/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifdef __PLM_MASTER_HEADER__
#error "plmsegment.h cannot be #included by another plastimatch header!"
#else
#define __PLM_MASTER_HEADER__


#ifndef _plmsegment_h_
#define _plmsegment_h_
#include "plmsegment_config.h"
#include "mabs.h"
#include "mabs_parms.h"
#include "mabs_subject.h"
#include "mabs_vote.h"
#endif /* #ifndef _plmsegment_h_ */


#undef __PLM_MASTER_HEADER__
#endif /* #ifdef __PLM_MASTER_HEADER__ */
