/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _drr_h_
#define _drr_h_

#include "plmreconstruct_config.h"

class Drr_options;

PLMRECONSTRUCT_API
void drr_compute (Drr_options *options);

#endif
