/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _plm_fwrite_h_
#define _plm_fwrite_h_

#include "plmsys_config.h"

PLMSYS_C_API void plm_fwrite (
        void* buf,
        size_t size,
        size_t count,
        FILE* fp, 
        bool force_little_endian
);

#endif

