/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _dcmtk_rdd_h_
#define _dcmtk_rdd_h_

#include "plmbase_config.h"
#include "rt_study_metadata.h"

void
dcmtk_load_rdd (
    Rt_study_metadata::Pointer rsd,
    const char *dicom_dir
);

#endif
