/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _vf_h_
#define _vf_h_

#include "plmbase_config.h"

class Volume;
class Volume_limit;

PLMBASE_C_API Volume* vf_warp (Volume* vout, Volume* vin, Volume* vf); 

#endif
