/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _opencl_probe_h_
#define _opencl_probe_h_

#include "plm_config.h"
#include "delayload.h"

#if defined __cplusplus
extern "C" {
#endif

plmopencl_EXPORT (
    int opencl_probe,
        void
);

#if defined __cplusplus
}
#endif

#endif
