/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _ray_trace_callback_h_
#define _ray_trace_callback_h_

#include "plmbase_config.h"

typedef void (*Ray_trace_callback) (
    void *callback_data, 
    size_t vox_index, 
    double vox_len, 
    float vox_value);

#endif
