/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _plm_return_code_h_
#define _plm_return_code_h_

#include "plmsys_config.h"

enum Plm_return_code
{
    PLM_UNINITIALIZED,    /* Do not use */
    PLM_SUCCESS,          /* General success */
    PLM_ERROR             /* Unspecified error */
};

#endif
