/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _translation_grid_search_h_
#define _translation_grid_search_h_

#include "plmregister_config.h"
#include "xform.h"

class Registration_data;
class Xform;
class Stage_parms;

Xform::Pointer
translation_grid_search_stage (
    Registration_data* regd,
    const Xform::Pointer& xf_in,
    const Stage_parms* stage);

#endif
