/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _xio_ct_h_
#define _xio_ct_h_

#include "plmbase_config.h"

class Plm_image;
class Xio_ct_transform;
class Xio_studyset;

PLMBASE_API void xio_ct_load (Plm_image *plm, Xio_studyset *xio_studyset);
PLMBASE_API void xio_ct_apply_transform (Plm_image *plm, Xio_ct_transform *transform);

#endif
