/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _autolabel_h_
#define _autolabel_h_

#include "plmsegment_config.h"

class Autolabel_parms;

PLMSEGMENT_API
void autolabel (Autolabel_parms *parms);

#endif
