/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _xform_point_h_
#define _xform_point_h_

#include "plmbase_config.h"
#include "itk_point.h"

class Xform;

PLMBASE_C_API void xform_point_transform (FloatPoint3DType* point_out, Xform* xf_in, FloatPoint3DType point_in);

#endif
