/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _itk_image_region_h_
#define _itk_image_region_h_

#include "itk_image.h"

template<class T> RegionType itk_image_region (const T&);
template<class T> RegionType itk_image_region (const T*);

#endif
