/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _float_pair_list_h_
#define _float_pair_list_h_

#include "plmbase_config.h"
#include <list>
#include <utility>
#include "itk_image_type.h"

typedef std::list< std::pair< float, float > > Float_pair_list;

PLMBASE_API Float_pair_list parse_float_pairs (const std::string& s);

#endif
