/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _pcmd_thumbnail_h_
#define _pcmd_thumbnail_h_

#include "plmcli_config.h"
#include <stdlib.h>
#include <string.h>

void
do_command_thumbnail (int argc, char *argv[]);

#endif
