/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _gabor_h_
#define _gabor_h_

#include "plmutil_config.h"
#include "plm_image.h"

class Gabor_private;

class Gabor 
{
public:
    Gabor_private *d_ptr;
public:
    Gabor ();
    ~Gabor ();

public:
    Plm_image::Pointer get_filter ();
};

#endif
