/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _plm_study_h_
#define _plm_study_h_

#include "plmbase_config.h"

class Plm_study_private;

class Plm_study 
{
public:
    Plm_study ();
    ~Plm_study ();

public:
    Plm_study_private *d_ptr;

public:
    void debug (void) const;
};

#endif
