/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _autolabel_feature_h_
#define _autolabel_feature_h_

#include "plmsegment_config.h"
#include <list>
#include <map>
#include <string>

class PLMSEGMENT_API Autolabel_feature {
public:
    Autolabel_feature ();
    ~Autolabel_feature ();

public:
    int feature_type;
    int gabor_uv[2];
    float gauss_width;
};

#endif
