/* -----------------------------------------------------------------------
   See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
   ----------------------------------------------------------------------- */
#ifndef _ramp_filter_h_
#define _ramp_filter_h_

#include "plmutil_config.h"

void ramp_filter (
    float *data, 
    unsigned int width, 
    unsigned int height
);

#endif
